----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT
-- Engineer:       Carlos Gil Soriano
--
-- Create Date:    00:47:34 10/26/2011
-- Design Name:    i2c bit recognition fsm
-- Module Name:    i2c_bit - Behavioral
-- Project Name:   Level Conversion Circuits
-- Target Devices: Spartan 6
-- Tool versions:
-- Description:    This is the fsm for the recognition of a bit. It offers some
--                 interrupting lines to the upper level (i2c_slave_core.vhd)
--                 which is responsible for managing byte level transactions
--
-- Dependencies:   The inputs should be debounced: i2c_debouncer.vhd
--
-- Revision:
-- Revision 0.01 - File Created
--          0.1  - Module works with no reported issues.
--          1.0  - Code revamped. Now is clearer and easier to read.
-- Additional Comments:
--
----------------------------------------------------------------------------------
library IEEE;
library work;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.i2c_slave_pkg.ALL;
use work.ctdah_pkg.ALL;

entity i2c_bit is
   port (rst_i     : in   STD_LOGIC;
         wb_clk_i  : in   STD_LOGIC;

         sda_i     : in   STD_LOGIC;
         scl_i     : in   STD_LOGIC;

         start_o   : out  STD_LOGIC;
         pause_o   : out  STD_LOGIC;
         rcved_o   : out  STD_LOGIC;
         done_o    : out  STD_LOGIC);
end i2c_bit;

architecture Behavioral of i2c_bit is


   type bit_fsm is (R0_RESET,
                    S0_IDLE,
                    S1A_HIGH_TMP,
                    S1A_HIGH,
                    S1B_LOW_TMP,
                    S1B_LOW,
                    S2A_START_TMP,
                    S2A_START,
                    S2B_STOP_DETECT,
                    Q1_ERROR);

   --! It specifies the maximum number of stages that will be employed for
   --! deglitching. Clocked with wb_clk_i
   constant c_MAX_GLITCH_DELAY : NATURAL := 6;
   --! Three delay stages out of six
   constant c_GLITCH_MASK      : STD_LOGIC_VECTOR (5 downto 0) := "000111";


   signal s_sda_deglitched     : STD_LOGIC;
   signal s_sda_deglitched_d1  : STD_LOGIC;

   signal s_scl_deglitched     : STD_LOGIC;
   signal s_scl_deglitched_d1  : STD_LOGIC;

   signal i2c_bit_fsm          : bit_fsm;

   signal s_scl_rising         : STD_LOGIC;
   signal s_scl_falling        : STD_LOGIC;

begin


   debouncer_scl_i: i2c_debouncer
   generic map(g_LENGTH   => 6)
   port map(rst          => rst_i,
            clk          => wb_clk_i,
            input        => scl_i,
            output       => s_scl_deglitched,
            glitch_mask  => c_GLITCH_MASK);

      --! Probably safer operation if we add one extra delay to the scl line.
      --! However, we increase the glitch time while placing an ACK.
      --! We have not implemented a counter to foresee the glitch due to the
      --! strange behaviour in Renesas I2C which rescales dinamically (that
      --! means in the middle of an I2C transaction) the scl line.
      --! Due to the variability of the scl period when an I2C transaction is
      --! in progress, it is better not to add a "guesser" of the scl period.

      ff1_scl : gc_ff
      port map(Q   => s_scl_deglitched_d1,
               C   => wb_clk_i,
               CLR => rst_i,
               D   => s_scl_deglitched);


   debouncer_sda_i: i2c_debouncer
   generic map(g_LENGTH   => 6)
   port map(rst          => rst_i,
            clk          => wb_clk_i,
            input        => sda_i,
            output       => s_sda_deglitched,
            glitch_mask  => c_GLITCH_MASK);

      ff1_sda : gc_ff
      port map(Q   => s_sda_deglitched_d1,
               C   => wb_clk_i,
               CLR => rst_i,
               D   => s_sda_deglitched);


   --! This is the process that samples the scl for detecting
   --! rise and falling edges
   reg_proc: process (wb_clk_i)
   begin
      if rising_edge(wb_clk_i) then
         if (s_scl_deglitched xor s_scl_deglitched_d1) = '1' then
            if s_scl_deglitched = '0' then
               s_scl_falling    <= '1';
            else
               s_scl_rising   <= '1';
            end if;
         else
            s_scl_rising       <= '0';
            s_scl_falling      <= '0';
         end if;
      else
      end if;
   end process;


   --! @brief Combiantion process to update the outputs.
   --! @param i2c_bit_fsm Finite state machine for the SDA bit detection.
   p_comb_output: process(i2c_bit_fsm)
   begin
      start_o     <= '0';
      pause_o     <= '0';
      rcved_o     <= '0';
      --! done_o is a one-clock signal
      done_o      <= '0';
      case i2c_bit_fsm is
         when  R0_RESET        =>
            null;
         when  S0_IDLE         =>
            null;
         when  S1A_HIGH   =>
            rcved_o     <= '1';
            done_o      <= '1';
         when  S1B_LOW    =>
            rcved_o     <= '0';
            done_o      <= '1';
         when  S2A_START  =>
            start_o     <= '1';
            done_o      <= '1';
         when  S2B_STOP_DETECT =>
            pause_o     <= '1';
            done_o      <= '1';
         when  Q1_ERROR   =>
            null;
         when  others     =>
            null;
      end case;
   end process;


   --! @brief  The fsm of this module, later on the sda sampled line is
   --!         validated in the falling edge of scl.
   --! @bparam wb_clk_i
   p_fsm: process(wb_clk_i)
   begin
      if rising_edge(wb_clk_i) then
         if rst_i = '1' then
            i2c_bit_fsm <= R0_RESET;
         --! After a detection of a falling edge we update the
         --! detection of a '0', a '1' and a start condition.
         elsif s_scl_falling = '1' then
            case i2c_bit_fsm is
               when S1A_HIGH_TMP  =>
                  i2c_bit_fsm <= S1A_HIGH;
               when S1B_LOW_TMP   =>
                  i2c_bit_fsm <= S1B_LOW;
               when S2A_START_TMP =>
                  i2c_bit_fsm <= S2A_START;
               when others =>
                  i2c_bit_fsm <= S0_IDLE;
            end case;

         --! When a rising edge is detected we annotate the first value
         --! in SDA: either a temporary '0' or '1'
         elsif s_scl_rising = '1' then
            if s_sda_deglitched_d1 = '1' then
               i2c_bit_fsm <= S1A_HIGH_TMP;
            else
               i2c_bit_fsm <= S1B_LOW_TMP;
            end if;

         else
            --! When we are in high level of a scl cycle, we keep on updating
            --! the fsm
            if s_scl_deglitched = '1' then
               case i2c_bit_fsm is
                  --! Just for random bit swapped coverage.
                  when S0_IDLE       =>
                     if s_sda_deglitched = '1' then
                        i2c_bit_fsm <= S1A_HIGH_TMP;
                     else
                        i2c_bit_fsm <= S1B_LOW_TMP;
                     end if;
                  when S1A_HIGH_TMP  =>
                     if s_sda_deglitched = '0' then
                     -- The detection of the start condition will be reported
                     -- in the next SCL rising edge.
                        i2c_bit_fsm <= S2A_START_TMP;
                     else
                     end if;
                  when S1B_LOW_TMP   =>
                     if s_sda_deglitched = '1' then
                     -- The detection of the pause condition MUST be
                     -- reported immediately.
                        i2c_bit_fsm <= S2B_STOP_DETECT;
                     else
                     end if;
                  when S2A_START_TMP =>
                     if s_sda_deglitched = '1' then
                        --! This happens if the deglitching is not enough
                        i2c_bit_fsm <= Q1_ERROR;
                     else
                     end if;
                  when others =>
                     i2c_bit_fsm <= S0_IDLE;
               end case;
            else
               if s_scl_deglitched_d1 = '0' then
                  i2c_bit_fsm <= S0_IDLE;
               else
               end if;
            end if;
         end if;
      else
      end if;
   end process;

end Behavioral;
