----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT
-- Engineer:       Carlos Gil Soriano
-- 
-- Create Date:    11:11:43 10/25/2011 
-- Design Name: 
-- Module Name:    i2c_debounce - Behavioral 
-- Project Name:   CTDAH
-- Target Devices: Spartan 6
-- Tool versions: 
-- Description:    This is a I2C debouncer. The main difference is that it is 
--                 intended to filter out low glitches. That means the oppositive
--                 from the previous version --debouncer.vhd.
--                 This behaviour makes more sense compared on how I2C definition
--                 resolves arbitration --conceptually the same as this module.
--
--                 A '1' in the glitch_mask means that the bit should be studied.
-- Dependencies:   None
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
library work;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.i2c_slave_pkg.ALL;
use work.ctdah_pkg.ALL;

entity i2c_debouncer is
   generic (g_LENGTH       : NATURAL := c_DEBOUNCE_LENGTH);
   port (rst         : in  STD_LOGIC;
         clk         : in  STD_LOGIC;
         input       : in  STD_LOGIC;
         output      : out STD_LOGIC;
         glitch_mask : in  STD_LOGIC_VECTOR(g_LENGTH - 1 downto 0));
end i2c_debouncer;

architecture Behavioral of i2c_debouncer is

   signal s_input_d0   : STD_LOGIC;
   --! The first of this signal is already stable (ff'ed two times at [0])
   signal s_delay      : STD_LOGIC_VECTOR(g_LENGTH - 1 downto 0);

begin

   ff1: gc_ff
   port map(
      Q    => s_input_d0,
      C    => clk,
      CLR  => rst,
      D    => input);

   ff2: gc_ff
   port map(
      Q    => s_delay(0),
      C    => clk,
      CLR  => rst,
      D    => s_input_d0);

   sync_delay_line: for i in 1 to g_LENGTH - 1 generate
      D_Flip_Flop : gc_ff
         port map (
            Q   => s_delay(i),
            C   => clk,
            CLR => rst,
            D   => s_delay(i-1));
    end generate sync_delay_line;

    p_output : process (clk)
    begin
       if rising_edge(clk) then
          if rst = '1' then
             output  <= '1';
          else
               --! We can deglitch either zeros or ones
             if (   (s_delay and glitch_mask)     = glitch_mask
             or (not(s_delay)and glitch_mask)     = glitch_mask) then
                output  <= s_delay(0);
             else
                  --! Internall pull-up of the pin
                output <= '1';
             end if;
          end if;
       end if;
    end process p_output;

end Behavioral;
