----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT
-- Engineer:       Carlos Gil Soriano
--
-- Create Date:    11:29:56 10/25/2011
-- Design Name:    I2C Slave to Wishbone bridge
-- Module Name:    i2c_slave_core - Behavioral
-- Project Name:   Level Conversion Circuits
-- Target Devices: Spartan 6
-- Tool versions:
-- Description:    This is a bridge from I2C to Wishbone.
--                 The module consist on the following interfaces:
--                    - I2C interface
--                    - Wishbone master
--                    - Wishbone slave
--                    - Interruption lines
--                 Each time a valid instruction is set in the I2C interface, this
--                 is reported either:
--                    A.- Via Wishbone: acting the module as a Wishbone master.
--                    B.- Interrupts:   the module offers some interrupt pins that
--                                      allows an external processor interact with
--                                      the module via the Wishbone slave
--                                      interface.
--                 The FSM of this I2C core is able to detect indirect I2C
--                 addressing. Thus reads, writes and prefetching of data
--                 through Wishbone master interface will be issued.
--                 ~ SUBCOMPONENTS ~
--                    This component evaluates the data at the byte level, upon
--                 the previous analysis of the SCL and SDA pins at a bit level.
--                 The bit level analysis is carried out in:
--                       --> ./i2c_bit.vhd
--                    Two internal FIFOs (one for RX and another TX) are combined
--                 with this module to store the received and ready-to-be sent
--                 frames. The files are:
--                       --> ./FIFO_stack.vhd
--                       --> ./FIFO_dispatcher.vhd
--                    Both the Wishbone master and slave interface access control
--                 and all configuration registers can be found in:
--                       --> ./i2c_regs.vhd
--
-- Dependencies:   i2c_bit.vhd, FIFO_stack.vhd, FIFO_dispatcher.vhd, gc_ff.vhd
--
-- Revision:
-- Revision 0.01 - File Created
--          0.1  - Initial submission. I2C instructions for Read and Write work.
--          1.0  - Code completely revamped. Code tested with a code coverage of
--                 81.5% with the testbench provided in:
--                       ../test/i2c_slave_test_top_tb.vhd
--
-- Additional Comments:
--
----------------------------------------------------------------------------------
library IEEE;
library work;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.i2c_slave_pkg.ALL;
use work.ctdah_pkg.ALL;

entity i2c_slave_core is
   generic(g_WB_CLK_PERIOD : TIME := 50 ns);
   port (
      clk           : in  STD_LOGIC;
      rst_i         : in  STD_LOGIC;
      -- I2C pins
      sda_oen       : out  STD_LOGIC;
      sda_i         : in   STD_LOGIC;
      sda_o         : out  STD_LOGIC;
      scl_oen       : out  STD_LOGIC;
      scl_i         : in   STD_LOGIC;
      scl_o         : out  STD_LOGIC;
      -- Registers directly accesible

      CTR0_i        : in  STD_LOGIC_VECTOR (r_CTR0'a_length - 1 downto 0);
      LT_o          : out STD_LOGIC_VECTOR (r_LT'a_length   - 1 downto 0);
      DRXA_o        : out STD_LOGIC_VECTOR (r_DRX'a_length  - 1 downto 0);
      DRXB_o        : out STD_LOGIC_VECTOR (r_DRX'a_length  - 1 downto 0);
      DTX_i         : in  STD_LOGIC_VECTOR (r_DTX'a_length  - 1 downto 0);
      load_TX       : in  STD_LOGIC;

      -- Alarms for controlling the i2c states
      pf_wb_addr_o  : out  STD_LOGIC;
      pf_wb_data_i  : in   STD_LOGIC_VECTOR(31 downto 0);
      rd_done_o     : out  STD_LOGIC;
      wr_done_o     : out  STD_LOGIC);
end i2c_slave_core;

architecture Behavioral of i2c_slave_core is

type SLA_fsm is (R0_RESET,
                 S0_IDLE,
                 S1_START_DETECT,
                 S2_I2C_ADDR,
                 S2A_I2C_ADDR_ACK,
                 S2NA_I2C_ADDR_NACK,
                 S3_WISHBONE_ADDR,
                 S3A_WISHBONE_ADDR_ACK,
                 S4_DETECT_OPERATION,
                 S5R_READ_SDA,
                 S5RA_READ_SDA_ACK,
                 S5W0_RESTART,
                 S5W1_I2C_ADDR,
                 S5W1A_I2C_ADDR_ACK,
                 S5W1NA_I2C_ADDR_NACK,
                 S5W2_WRITE_SDA,
                 S5W2A_WRITE_SDA_ACK,
                 S5W3_NACK,
                 S6_WAIT_START_PAUSE,
                 S7_PAUSE_DETECT);

-------------------------------------------------------------------------------
-- FSM signals for the I2C module
-------------------------------------------------------------------------------
-- Four signals are employed to govern the FSM behaviour and its outputs.
--
-- These signals are updated in done_comb_proc process, dependant of the done
-- signal coming from i2c_bit.vhd
-- Every time a deglitched falling edge of the SCL line is detected in the
-- aforementioned i2c_bit.vhd, the signals done, start_o, pause_o, rcved_o are
-- considered valid to be studied in i2c_slave_core.vhd processes.
--
-------------------------------------------------------------------------------


constant c_WATCHDOG_END_VALUE : NATURAL := c_WATCHDOG_DEADLINE/g_WB_CLK_PERIOD;

signal i2c_SLA_fsm            : SLA_fsm := R0_RESET;
signal i2c_SLA_fsm_d0         : SLA_fsm := R0_RESET;

signal s_DRXA_slv  : STD_LOGIC_VECTOR(r_DRX'a_length - 1 downto 0)
                                            := (others => '0');
signal s_DRXB_slv  : STD_LOGIC_VECTOR(r_DRX'a_length - 1 downto 0)
                                            := (others => '0');
signal s_DRX_slv   : STD_LOGIC_VECTOR(r_DRX'a_length*2 - 1 downto 0)
                                            := (others => '0');
signal s_DTX_slv   : STD_LOGIC_VECTOR(r_DTX'a_length - 1 downto 0);

signal s_CTR0      : r_CTR0;
signal s_LT        : r_LT  := c_LT_default;

signal s_i2c_addr_ctr0  : STD_LOGIC_VECTOR(7 downto 0);
signal s_i2c_addr       : STD_LOGIC_VECTOR(7 downto 0);

-------------------------------------------------------------------------------
-- i2c signals
-------------------------------------------------------------------------------
signal s_sda_o            : STD_LOGIC;
signal s_sda_oen          : STD_LOGIC;

signal s_bit_done         : STD_LOGIC;

-- Bit counter signals
signal s_bit_cnt_slv      : STD_LOGIC_VECTOR(c_COUNTER_WIDTH - 1 downto 0);
signal s_bit_cnt          : NATURAL;
signal s_bit_cnt_rst      : STD_LOGIC;

-- Byte counter signals
signal s_byte_cnt_slv     : STD_LOGIC_VECTOR(c_COUNTER_WIDTH - 1 downto 0);
signal s_byte_cnt         : NATURAL;
signal s_byte_cnt_rst     : STD_LOGIC;
signal s_byte_cnt_en      : STD_LOGIC := '0';

-- TX FIFO (FIFO_dispatcher) signals
signal s_tx_fifo_load_TX  : STD_LOGIC;
signal s_tx_fifo_flush    : STD_LOGIC;
signal s_tx_fifo_oen      : STD_LOGIC;
signal s_tx_fifo_reg_o    : STD_LOGIC_VECTOR (c_DATA_WIDTH -1 downto 0);

-- Interrupting signals (i2c_bit.vhd)
signal s_start_o          : STD_LOGIC;
signal s_pause_o          : STD_LOGIC;
signal s_rcved_o          : STD_LOGIC;

signal s_pf_wb_data       : STD_LOGIC_VECTOR(31 downto 0) := (others => '1');

signal s_watchdog_cnt_rst : STD_LOGIC;
signal s_watchdog_cnt_slv : STD_LOGIC_VECTOR(c_WATCHDOG_WIDTH - 1 downto 0);
signal s_watchdog_cnt     : NATURAL;

begin

   s_CTR0       <= f_CTR0(CTR0_i);
   LT_o         <= f_STD_LOGIC_VECTOR(s_LT);
   s_pf_wb_data <= pf_wb_data_i;

   inst_i2c_bit  : i2c_bit
   port map(rst_i     => rst_i,
            wb_clk_i  => clk,
            sda_i     => sda_i,
            scl_i     => scl_i,
            start_o   => s_start_o,
            pause_o   => s_pause_o,
            rcved_o   => s_rcved_o,
            done_o    => s_bit_done);


   bit_counter_8: gc_counter
   generic map (g_data_width => c_DATA_WIDTH)
   port map (clk_i  => clk,
             rst_i  => s_bit_cnt_rst,
             en_i   => s_bit_done,
             cnt_o  => s_bit_cnt_slv);

      s_bit_cnt  <= to_integer(UNSIGNED(s_bit_cnt_slv));

   byte_counter_8: gc_counter
   generic map (g_data_width => c_DATA_WIDTH)
   port map (clk_i  => clk,
             rst_i  => s_byte_cnt_rst,
             en_i   => s_byte_cnt_en,
             cnt_o  => s_byte_cnt_slv);

      s_byte_cnt <= to_integer(UNSIGNED(s_byte_cnt_slv));

   watchdog_counter_8: gc_counter
   generic map (g_data_width => c_WATCHDOG_WIDTH)
   port map (clk_i  => clk,
             rst_i  => s_watchdog_cnt_rst,
             en_i   => c_WATCHDOG_ENABLE,
             cnt_o  => s_watchdog_cnt_slv);

      s_watchdog_cnt       <= to_integer(UNSIGNED(s_watchdog_cnt_slv));



   scl_o   <= '1';
   scl_oen <= '0';

      sda_o <= s_sda_o;

   p_sda_o: process(i2c_SLA_fsm,
                    s_bit_cnt_slv(0))
      variable v_bit_inv    : UNSIGNED(2 downto 0);
      variable v_bit_order  : NATURAL;
   begin
         case i2c_SLA_fsm is
            when S2A_I2C_ADDR_ACK      =>
               s_sda_o <= '0';
            when S3A_WISHBONE_ADDR_ACK =>
               s_sda_o <= '0';
            when S5RA_READ_SDA_ACK     =>
               s_sda_o <= '0';
            when S5W1A_I2C_ADDR_ACK    =>
               s_sda_o <= '0';
            when S5W2_WRITE_SDA        =>
               v_bit_inv   := UNSIGNED(not(s_bit_cnt_slv(2 downto 0)));
               v_bit_order := to_integer(v_bit_inv);
               s_sda_o <= s_pf_wb_data(s_byte_cnt*8 + v_bit_order);
            when others           =>
               s_sda_o <= '1';
         end case;
   end process;

      sda_oen <= s_sda_oen;

   p_sda_oen: process(i2c_SLA_fsm)
   begin
      s_sda_oen <= '0';
      case i2c_SLA_fsm is
         when S2A_I2C_ADDR_ACK      =>
            s_sda_oen <= '1';
         when S3A_WISHBONE_ADDR_ACK =>
            s_sda_oen <= '1';
         when S5RA_READ_SDA_ACK     =>
            s_sda_oen <= '1';
         when S5W1A_I2C_ADDR_ACK    =>
            s_sda_oen <= '1';
         when S5W2_WRITE_SDA        =>
            s_sda_oen <= '1';
         when others                =>
            null;
      end case;
   end process;

   s_DRXA_slv <= s_DRX_slv(r_DRX'a_length - 1    downto 0);
   s_DRXB_slv <= s_DRX_slv(r_DRX'a_length*2 - 1 downto r_DRX'a_length);

   --! @brief      Process to store the incoming data
   --! @param clk  Clock that samples the data
   p_store_FIFO: process(clk)

      procedure shift_DRX is
      begin
         s_DRX_slv(s_DRX_slv'length - 1) <= s_rcved_o;
         for i in 1 to s_DRX_slv'length - 1 loop
            s_DRX_slv(i-1) <= s_DRX_slv(i);
         end loop;
      end procedure;

   begin
      if rising_edge(clk) then
         if s_bit_done = '1' then
            case i2c_SLA_fsm is
               when S2_I2C_ADDR      =>
                  shift_DRX;
               when S3_WISHBONE_ADDR =>
                  shift_DRX;
               when S4_DETECT_OPERATION =>
                  if     s_start_o = '0'
                     and s_pause_o = '0' then
                        shift_DRX;
                  end if;
               when S5R_READ_SDA     =>
                  shift_DRX;
               when S5W1_I2C_ADDR    =>
                  shift_DRX;
               when others =>
                  null;
            end case;
         end if;
      end if;
   end process;


   s_i2c_addr_ctr0(7 downto 1) <= STD_LOGIC_VECTOR(s_CTR0.I2C_ADDR);
   s_i2c_addr(0)               <= '0';

   gen_i2c: for i in 1 to 7 generate
      s_i2c_addr(8-i) <= s_i2c_addr_ctr0(i);
   end generate gen_i2c;


   p_delays : process(clk)
   begin
      if rising_edge(clk) then
         i2c_SLA_fsm_d0   <= i2c_SLA_fsm;
      end if;
   end process;


   --! @brief     Process to generate the pf_wb_addr_o signal.
   --! @param clk Main clock
   p_pf_wb_addr: process(clk)
   begin
      if rising_edge(clk) then
         pf_wb_addr_o <= '0';
         if     i2c_SLA_fsm    = S5W1A_I2C_ADDR_ACK
            and i2c_SLA_fsm_d0 = S5W1_I2C_ADDR      then
            pf_wb_addr_o <= '1';
         end if;
      end if;
   end process;

   --! @brief     Process to generate the rd_done_o signal.
   --! @param clk Main clock
   p_rd_done: process(clk)
   begin
      if rising_edge(clk) then
         rd_done_o <= '0';
         if     i2c_SLA_fsm = S6_WAIT_START_PAUSE
            and i2c_SLA_fsm_d0 = S5RA_READ_SDA_ACK     then
            rd_done_o <= '1';
         end if;
      end if;
   end process;

   --! @brief     Process to generate the wr_done_o signal.
   --! @param clk Main clock
   p_wr_done: process(clk)
   begin
      if rising_edge(clk) then
         wr_done_o <= '0';
         if     i2c_SLA_fsm    = S6_WAIT_START_PAUSE
            and i2c_SLA_fsm_d0 = S5W3_NACK           then
            wr_done_o <= '1';
         end if;
      end if;
   end process;

   DRXA_o   <= s_DRXA_slv;
   DRXB_o   <= s_DRXB_slv;

--   tx_fifo_i2c: FIFO_dispatcher
--   generic map(g_data_width       => c_DATA_WIDTH,
--               g_dispatcher_depth => c_TX_STACK_DEPTH)
--   port map(reg_i(7  downto  0)  => s_DTX_slv(31 downto 24),
--            reg_i(15 downto  8)  => s_DTX_slv(23 downto 16),
--            reg_i(23 downto 16)  => s_DTX_slv(15 downto 8),
--            reg_i(31 downto 24)  => s_DTX_slv(7  downto 0),
--            clk                  => clk,
--            load                 => s_tx_fifo_load_TX,
--            flush                => s_tx_fifo_flush,
--            oen_i                => s_tx_fifo_oen,
--            reg_o                => s_tx_fifo_reg_o);

   --! @brief process to update the signals that drive bit_counter_8
   --! @param i2c_SLA_fsm    Current value of the i2c fsm
   --! @param i2c_SLA_fsm_d0 Current value of the i2c fsm
   p_bit_counter_comb : process(i2c_SLA_fsm,
                                i2c_SLA_fsm_d0,
                                s_byte_cnt,
                                s_CTR0.BIA)
   begin
      s_bit_cnt_rst  <= '0';
      s_byte_cnt_rst <= '0';
      case i2c_SLA_fsm is
         when R0_RESET               =>
            if i2c_SLA_fsm_d0 /= R0_RESET then
               s_bit_cnt_rst     <= '1';
               s_byte_cnt_rst    <= '1';
            end if;
         when S0_IDLE                =>
            null;
         when S1_START_DETECT        =>
            s_bit_cnt_rst     <= '1';
            s_byte_cnt_rst    <= '1';
         when S2_I2C_ADDR            =>
            null;
         when S2A_I2C_ADDR_ACK       =>
            if i2c_SLA_fsm_d0 = S2_I2C_ADDR then
               s_byte_cnt_rst <= '1';
            end if;
         when S3_WISHBONE_ADDR       =>
            if    i2c_SLA_fsm_d0 = S2A_I2C_ADDR_ACK
               or i2c_SLA_fsm_d0 = S3A_WISHBONE_ADDR_ACK
               or i2c_SLA_fsm_d0 = S5W1A_I2C_ADDR_ACK    then
               s_bit_cnt_rst  <= '1';
            end if;
            if s_byte_cnt = s_CTR0.BIA then
               if i2c_SLA_fsm_d0 = S2A_I2C_ADDR_ACK then
                  s_byte_cnt_rst <= '1';
               end if;
            end if;
         when S3A_WISHBONE_ADDR_ACK  =>
            if i2c_SLA_fsm_d0 = S3_WISHBONE_ADDR then
               s_bit_cnt_rst <= '1';
            end if;
         when S4_DETECT_OPERATION    =>
            if i2c_SLA_fsm_d0 = S3A_WISHBONE_ADDR_ACK then
               s_bit_cnt_rst  <= '1';
               s_byte_cnt_rst <= '1';
            end if;
         when S5W1_I2C_ADDR          =>
            null;
         when S5W1A_I2C_ADDR_ACK     =>
            if i2c_SLA_fsm_d0 <= S5W1_I2C_ADDR then
               s_bit_cnt_rst <= '1';
            end if;
         when S5W2_WRITE_SDA         =>
            if s_byte_cnt = s_CTR0.BIA then
               if i2c_SLA_fsm_d0 = S3A_WISHBONE_ADDR_ACK then
                  s_byte_cnt_rst <= '1';
               end if;
            end if;
            if    i2c_SLA_fsm_d0 = S5W1A_I2C_ADDR_ACK
               or i2c_SLA_fsm_d0 = S5W2A_WRITE_SDA_ACK then
               s_bit_cnt_rst <= '1';
            end if;
         when S5W2A_WRITE_SDA_ACK    =>
            if i2c_SLA_fsm_d0 = S5W2_WRITE_SDA then
               s_bit_cnt_rst <= '1';
            end if;
         when S5W0_RESTART           =>
            if i2c_SLA_fsm_d0 = S4_DETECT_OPERATION then
               s_bit_cnt_rst     <= '1';
            end if;
         when S5R_READ_SDA           =>
            if s_byte_cnt = s_CTR0.BIA then
               if    i2c_SLA_fsm_d0 = S3A_WISHBONE_ADDR_ACK then
                  s_byte_cnt_rst <= '1';
               end if;
            end if;
            if i2c_SLA_fsm_d0 = S5RA_READ_SDA_ACK          then
               s_bit_cnt_rst     <= '1';
            end if;
         when S5RA_READ_SDA_ACK      =>
            if i2c_SLA_fsm_d0 = S5R_READ_SDA then
               s_bit_cnt_rst <= '1';
            end if;
         when S6_WAIT_START_PAUSE    =>
            if i2c_SLA_fsm_d0 /= S6_WAIT_START_PAUSE then
                  s_bit_cnt_rst  <= '1';
                  s_byte_cnt_rst <= '1';
            end if;
         when S7_PAUSE_DETECT        =>
            null;
         when others                 =>
               s_bit_cnt_rst  <= '1';
               s_byte_cnt_rst <= '1';
      end case;
   end process;

   p_byte_counter_en: process (i2c_SLA_fsm,
                               i2c_SLA_fsm_d0,
                               s_CTR0.BIA)
   begin
      s_byte_cnt_en <= '0';
      case i2c_SLA_fsm is
         when R0_RESET         =>
            if i2c_SLA_fsm_d0 /= R0_RESET then
            end if;
         when S3A_WISHBONE_ADDR_ACK =>
            if i2c_SLA_fsm_d0 = S3_WISHBONE_ADDR then
               s_byte_cnt_en <= '1';
            end if;
         when S5RA_READ_SDA_ACK     =>
            if i2c_SLA_fsm_d0 = S5R_READ_SDA then
               s_byte_cnt_en <= '1';
            end if;
         when S5W2A_WRITE_SDA_ACK   =>
            if i2c_SLA_fsm_d0 = S5W2_WRITE_SDA then
               s_byte_cnt_en <= '1';
            end if;
         when others           =>
            null;
      end case;
   end process;

   --! @brief       Watchdog reset process
   --! @param clk   Main clock
   p_watchdog_rst: process(clk)
   begin
      if rising_edge(clk) then
         s_watchdog_cnt_rst <= '0';
         if rst_i = '1' then
            s_watchdog_cnt_rst <= '1';
         end if;
         if s_watchdog_cnt >= c_WATCHDOG_END_VALUE then
            s_watchdog_cnt_rst <= '1';
         end if;
         if i2c_SLA_fsm /= i2c_SLA_fsm_d0 then
            s_watchdog_cnt_rst <= '1';
         end if;
      end if;
   end process;


   --! @brief                Small process that records the length of a
   --!                       bit over the I2C interface.
   --! @param i2c_SLA_fsm    Actual value of the fsm
   --! @param i2c_SLA_fsm_d0 One-wb_clocked delayed value of the fsm
   p_LT_SCLP: process(i2c_SLA_fsm,
                      i2c_SLA_fsm_d0)
                      variable v_count  : STD_LOGIC_VECTOR(23 downto 0);
   begin
      if     i2c_SLA_fsm    = S2A_I2C_ADDR_ACK
         and i2c_SLA_fsm_d0 = S2_I2C_ADDR       then
            --! Here we do a division bit 8, because the watchdog timer,
            --! is counting for 8 scl clocks
            v_count := s_watchdog_cnt_slv(24 - 1 + 3 downto 3);
            s_LT.SCLP <= UNSIGNED(v_count);
      end if;
   end process;

   --! @brief      Process to set the next fsm
   --! @param clk  Clock that rules the fsm transitions
   fsm_proc: process(clk)

      procedure check_start_stop is
      begin
         if    s_start_o = '1' then
            i2c_SLA_fsm <= S1_START_DETECT;
         elsif s_pause_o = '1' then
            i2c_SLA_fsm <= R0_RESET;
         end if;
      end procedure;

   begin
      if rising_edge(clk) then
         if s_watchdog_cnt >= c_WATCHDOG_END_VALUE then
            i2c_SLA_fsm <= R0_RESET;
         else
            case i2c_SLA_fsm is
               when R0_RESET =>
                  i2c_SLA_fsm <= S0_IDLE;
               when S0_IDLE               =>
                  if s_bit_done = '1' then
                     if s_start_o = '1' then
                        i2c_SLA_fsm <= S1_START_DETECT;
                     else
                        i2c_SLA_fsm <= R0_RESET;
                     end if;
                  end if;
               when S1_START_DETECT       =>
                  i2c_SLA_fsm <= S2_I2C_ADDR;
               when S2_I2C_ADDR           =>
                  if s_bit_done = '1' then
                     if s_bit_cnt = 7 then
                        if   s_DRX_slv(s_DRX_slv'length - 1 downto
                                       s_DRX_slv'length - 1 - 6) = "0000000"
                          or s_DRX_slv(s_DRX_slv'length - 1 downto
                                       s_DRX_slv'length - 1 - 6) =
                                                       s_i2c_addr(7 downto 1)  then
                           i2c_SLA_fsm    <= S2A_I2C_ADDR_ACK;
                        else
                           i2c_SLA_fsm    <= S2NA_I2C_ADDR_NACK;
                        end if;
                     end if;
                     check_start_stop;
                  end if;
               when S2A_I2C_ADDR_ACK      =>
                  if s_bit_done = '1' then
                     i2c_SLA_fsm        <= S3_WISHBONE_ADDR;
                     check_start_stop;
                  end if;
               when S2NA_I2C_ADDR_NACK    =>
                  if s_bit_done = '1' then
                     i2c_SLA_fsm        <= S6_WAIT_START_PAUSE;
                     check_start_stop;
                  end if;
               when S3_WISHBONE_ADDR      =>
                  if s_bit_done = '1' then
                     if s_bit_cnt = 7 then
                        i2c_SLA_fsm <= S3A_WISHBONE_ADDR_ACK;
                     end if;
                     check_start_stop;
                  end if;
               when S3A_WISHBONE_ADDR_ACK =>
                  if s_bit_done = '1' then
                     if s_byte_cnt < s_CTR0.BIA then
                        i2c_SLA_fsm <= S3_WISHBONE_ADDR;
                     else
                        i2c_SLA_fsm <= S4_DETECT_OPERATION;
                        check_start_stop;
                     end if;
                  end if;
               when S4_DETECT_OPERATION    =>
                  if s_bit_done = '1' then
                     if s_start_o = '1' then
                        i2c_SLA_fsm <= S5W0_RESTART;
                     elsif s_pause_o = '1' then
                        i2c_SLA_fsm <= R0_RESET;
                     else
                        i2c_SLA_fsm <= S5R_READ_SDA;
                        check_start_stop;
                     end if;
                  end if;
               when S5R_READ_SDA           =>
                  if s_bit_done = '1' then
                     if s_bit_cnt = 7 then
                        i2c_SLA_fsm <= S5RA_READ_SDA_ACK;
                     end if;
                     check_start_stop;
                  end if;
               when S5RA_READ_SDA_ACK      =>
                  if s_bit_done = '1' then
                     if s_byte_cnt < s_CTR0.BRD then
                        i2c_SLA_fsm <= S5R_READ_SDA;
                     else
                        i2c_SLA_fsm <= S6_WAIT_START_PAUSE;
                        check_start_stop;
                     end if;
                  end if;
               when S5W0_RESTART            =>
                  i2c_SLA_fsm <= S5W1_I2C_ADDR;
               when S5W1_I2C_ADDR           =>
                  if s_bit_done = '1' then
                     if s_bit_cnt = 7 then
                        if   s_DRX_slv(s_DRX_slv'length - 1 downto
                                       s_DRX_slv'length - 1 - 6) = "0000000"
                          or s_DRX_slv(s_DRX_slv'length - 1 downto
                                       s_DRX_slv'length - 1 - 6) =
                                                       s_i2c_addr(7 downto 1)  then
                           i2c_SLA_fsm  <= S5W1A_I2C_ADDR_ACK;
                        else
                           i2c_SLA_fsm  <= S5W1NA_I2C_ADDR_NACK;
                        end if;
                     end if;
                     check_start_stop;
                  end if;
               when S5W1A_I2C_ADDR_ACK      =>
                  if s_bit_done = '1' then
                     i2c_SLA_fsm        <= S5W2_WRITE_SDA;
                     check_start_stop;
                  end if;
               when S5W1NA_I2C_ADDR_NACK    =>
                  if s_bit_done = '1' then
                     i2c_SLA_fsm        <= S6_WAIT_START_PAUSE;
                     check_start_stop;
                  end if;
               when S5W2_WRITE_SDA          =>
                  if s_bit_done = '1' then
                     if s_bit_cnt = 7 then
                        i2c_SLA_fsm     <= S5W2A_WRITE_SDA_ACK;
                     end if;
                     check_start_stop;
                  end if;
               when S5W2A_WRITE_SDA_ACK     =>
                  if s_bit_done = '1' then
                     if s_rcved_o = '0' then
                        if s_byte_cnt < s_CTR0.BWR then
                           i2c_SLA_fsm <= S5W2_WRITE_SDA;
                        else
                           i2c_SLA_fsm <= R0_RESET;
                        end if;
                     else
                        i2c_SLA_fsm <= S5W3_NACK;
                        check_start_stop;
                     end if;
                  end if;
               when S5W3_NACK             =>
                  i2c_SLA_fsm <= S6_WAIT_START_PAUSE;
               when S6_WAIT_START_PAUSE   =>
                  if s_bit_done = '1' then
                     if s_pause_o = '1' then
                        i2c_SLA_fsm <= S7_PAUSE_DETECT;
                     elsif s_start_o = '1' then
                        i2c_SLA_fsm <= S1_START_DETECT;
                     else
                        i2c_SLA_fsm <= R0_RESET;
                        check_start_stop;
                     end if;
                  else
                     --!TODO Implement here a watchdog
                  end if;
               when S7_PAUSE_DETECT       =>
                  i2c_SLA_fsm <= R0_RESET;
                  check_start_stop;
               when others                =>
                  i2c_SLA_fsm <= R0_RESET;
            end case;
         end if;
      end if;
   end process;

end Behavioral;
