----------------------------------------------------------------------------------
--
-- Company:        CERN, BE-CO-HT
-- Engineer:       Carlos Gil Soriano
--
-- Create Date:    11:03:11 06/19/2012
-- Design Name:
-- Module Name:    i2c_slave_pkg - Behavioral
-- Project Name:   CTDAH/SVEC
-- Target Devices: Spartan 6
-- Tool versions:  Xilinx ISE 13.4
-- Description:    Package for the i2c_slave
--
-- Dependencies:
--
-- Revision: 0.1
--   0.01          + File Created
-- Additional Comments:
--
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.ALL;

package i2c_slave_pkg is

   --! Default clock divider
   constant c_RST_EXTENSOR       : NATURAL := 4;
   constant c_DEBOUNCE_LENGTH    : NATURAL := 6;
   --! Broadcast address as specified in I2C standard
   constant c_I2C_GENERAL_ADDR   : UNSIGNED(6 downto 0) := to_unsigned(0,7);
   constant c_RETRY_LENGTH       : NATURAL := 10; --! 10 clks max. retries

   constant c_SDA_reverse_bytes  : BOOLEAN := false;
   constant c_SDA_MSb            : BOOLEAN := false;

   -------------------------------------------------------------------------------
   -- This is the declaration area of the constants used in other components:
   -- c_COUNTER_WIDTH    bit length of the gc_counter
   -- c_DATA_WIDTH       width of the data to be stored in FIFO_stack
   --                    and FIFO_dispatchers.
   -- c_RX_STACK_DEPTH   depth of FIFO_stack
   -- c_TX_STACK_DEPTH   depth of FIFO_dispatch
   -------------------------------------------------------------------------------
   constant c_COUNTER_WIDTH     : NATURAL := 8;
   constant c_DATA_WIDTH        : NATURAL := 8;
   constant c_RX_STACK_DEPTH    : NATURAL := 6;
   constant c_TX_STACK_DEPTH    : NATURAL := 4;

   constant c_WB_CLK_PERIOD     : TIME      := 50 ns;
   constant c_WATCHDOG_WIDTH    : NATURAL   := 32;
   constant c_WATCHDOG_DEADLINE : TIME      := 10**9*60 ns;
   constant c_WATCHDOG_ENABLE   : STD_LOGIC := '1';
   attribute a_length : NATURAL;


   -------------------------------------
   -- CTR0 register
   -------------------------------------
   -- 32 bits
   -- Wishbone access: Write-read
   ------------------------------------
   -- BIT       NAME           Description
   --     0      I2C_OP         I2C OPeration
   --   7-1      I2C_ADDR       I2C address
   --  11-8      BIA            Bytes of Indirect Addressing
   --  19-12     BRD            Bytes to be ReaD from FPGA
   --  27-20     BWR            Bytes to be WRitten to FPGA
   --  31-28     x              Reserved
   -------------------------------------
   --  I2C_OP    Operations
   --       0    Read from I2C
   --       1    Write into I2C
   -------------------------------------
   type r_CTR0 is
     record
            I2C_OP    : STD_LOGIC;
            I2C_ADDR  : UNSIGNED( 7 downto  1);
            BIA       : UNSIGNED(11 downto  8);
            BRD       : UNSIGNED(19 downto 12);
            BWR       : UNSIGNED(27 downto 20);
            x         : STD_LOGIC_VECTOR(31 downto 28);
     end record;

   attribute a_length of r_CTR0 : type is 32;

   -------------------------------------
   -- LT register
   -------------------------------------
   -- 32 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --   7-0      WBCP           WishBone Clock Period
   --  31-8      SCLP           SCL Period
   -------------------------------------
   -- Both period are given with nanosecond
   -- precision.
   -------------------------------------
   type r_LT is
      record
         WBCP : UNSIGNED(7  downto 0);
         SCLP : UNSIGNED(31 downto 8);
      end record;

   attribute a_length of r_LT : type is 32;

   -------------------------------------
   -- DRX register
   -------------------------------------
   -- 32 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --  31-0      DRX            Data RX register
   -------------------------------------
   type r_DRX is
     record
        data : STD_LOGIC_VECTOR(31 downto 0);
     end record;

   attribute a_length of r_DRX : type is 32;

   -------------------------------------
   -- DTX register
   -------------------------------------
   -- 32 bits
   -- Wishbone access: Read-only
   ------------------------------------
   -- BIT       NAME           Description
   --  31-0      DTX            Data TX register
   -------------------------------------
   type r_DTX is
     record
        data : STD_LOGIC_VECTOR(31 downto 0);
     end record;

   attribute a_length of r_DTX : type is 32;


   constant c_CTR0_addr     : UNSIGNED(3 downto 0) := X"0";
   constant c_LT_addr       : UNSIGNED(3 downto 0) := X"1";
   constant c_DTX_addr      : UNSIGNED(3 downto 0) := X"2";
   constant c_DRXA_addr     : UNSIGNED(3 downto 0) := X"3";
   constant c_DRXB_addr     : UNSIGNED(3 downto 0) := X"4";

   constant c_CTR0_default : r_CTR0 := (I2C_OP    => '0',
                                        I2C_ADDR  => c_I2C_GENERAL_ADDR,
                                        BIA       => to_unsigned(2, 4),
                                        BRD       => to_unsigned(4, 8),
                                        BWR       => to_unsigned(4, 8),
                                        x         => (others => '0'));

   constant c_LT_default   : r_LT := (WBCP =>
                                       to_unsigned(c_WB_CLK_PERIOD/(1 ns), 8),
                                      SCLP => to_unsigned(0, 24));

   component i2c_slave_core
   generic(g_WB_CLK_PERIOD : TIME := 50 ns);
   port (
      clk          : in   STD_LOGIC;
      rst_i        : in   STD_LOGIC;

      sda_oen      : out  STD_LOGIC;
      sda_i        : in   STD_LOGIC;
      sda_o        : out  STD_LOGIC;
      scl_oen      : out  STD_LOGIC;
      scl_i        : in   STD_LOGIC;
      scl_o        : out  STD_LOGIC;

      CTR0_i       : in  STD_LOGIC_VECTOR (r_CTR0'a_length - 1 downto 0);
      LT_o         : out STD_LOGIC_VECTOR (r_LT'a_length   - 1 downto 0);
      DRXA_o       : out STD_LOGIC_VECTOR (r_DRX'a_length  - 1 downto 0);
      DRXB_o       : out STD_LOGIC_VECTOR (r_DRX'a_length  - 1 downto 0);
      DTX_i        : in  STD_LOGIC_VECTOR (r_DTX'a_length  - 1 downto 0);

      load_TX      : in   STD_LOGIC;
      pf_wb_addr_o : out  STD_LOGIC;
      pf_wb_data_i : in  STD_LOGIC_VECTOR(31 downto 0);
      rd_done_o    : out  STD_LOGIC;
      wr_done_o    : out  STD_LOGIC);
   end component;

   component i2c_regs
      port (wb_clk           : in  STD_LOGIC;
            wb_rst_i         : in  STD_LOGIC;

            wb_master_we_o   : out STD_LOGIC;
            wb_master_stb_o  : out STD_LOGIC;
            wb_master_cyc_o  : out STD_LOGIC;
            wb_master_sel_o  : out STD_LOGIC_VECTOR (3  downto 0);
            wb_master_data_i : in  STD_LOGIC_VECTOR (31 downto 0);
            wb_master_data_o : out STD_LOGIC_VECTOR (31 downto 0);
            wb_master_addr_o : out STD_LOGIC_VECTOR (15 downto 0);
            wb_master_ack_i  : in  STD_LOGIC;
            wb_master_rty_i  : in  STD_LOGIC;
            wb_master_err_i  : in  STD_LOGIC;

            wb_slave_we_i    : in  STD_LOGIC;
            wb_slave_stb_i   : in  STD_LOGIC;
            wb_slave_cyc_i   : in  STD_LOGIC;
            wb_slave_sel_i   : in  STD_LOGIC_VECTOR(3  downto 0);
            wb_slave_data_i  : in  STD_LOGIC_VECTOR(31 downto 0);
            wb_slave_data_o  : out STD_LOGIC_VECTOR(31 downto 0);
            wb_slave_addr_i  : in  STD_LOGIC_VECTOR(3  downto 0);
            wb_slave_ack_o   : out STD_LOGIC;
            wb_slave_rty_o   : out STD_LOGIC;
            wb_slave_err_o   : out STD_LOGIC;

            CTR0_o           : out STD_LOGIC_VECTOR (r_CTR0'a_length - 1 downto 0);
            LT_i             : in  STD_LOGIC_VECTOR (r_LT'a_length   - 1 downto 0);
            DRXA_i           : in  STD_LOGIC_VECTOR (r_DRX'a_length  - 1 downto 0);
            DRXB_i           : in  STD_LOGIC_VECTOR (r_DRX'a_length  - 1 downto 0);
            DTX_o            : out STD_LOGIC_VECTOR (r_DTX'a_length  - 1 downto 0);
            load_TX          : out STD_LOGIC;

            pf_wb_addr_i     : in  STD_LOGIC;
            pf_wb_data_o     : out STD_LOGIC_VECTOR(31 downto 0);
            rd_done_i        : in  STD_LOGIC;
            wr_done_i        : in  STD_LOGIC;
            i2c_addr_i       : in  STD_LOGIC_VECTOR(6 downto 0));
   end component;

   component i2c_debouncer is
      generic (g_LENGTH       : NATURAL := c_DEBOUNCE_LENGTH);
      port (rst         : in  STD_LOGIC;
            clk         : in  STD_LOGIC;
            input       : in  STD_LOGIC;
            output      : out STD_LOGIC;
            glitch_mask : in  STD_LOGIC_VECTOR(g_LENGTH - 1 downto 0));
   end component;

   component i2c_bit
      port (rst_i     : in   STD_LOGIC;
            wb_clk_i  : in   STD_LOGIC;
            sda_i     : in   STD_LOGIC;
            scl_i     : in   STD_LOGIC;
            start_o   : out  STD_LOGIC;
            pause_o   : out  STD_LOGIC;
            rcved_o   : out  STD_LOGIC;
            done_o    : out  STD_LOGIC);
   end component;

   function f_CTR0  (signal r_register : in STD_LOGIC_VECTOR(31 downto 0))
                                                                 return r_CTR0;
   function f_LT    (signal r_register : in STD_LOGIC_VECTOR(31 downto 0))
                                                                 return r_LT;


   function f_STD_LOGIC_VECTOR(r_register : in r_CTR0) return STD_LOGIC_VECTOR;
   function f_STD_LOGIC_VECTOR(r_register : in r_LT)   return STD_LOGIC_VECTOR;

end i2c_slave_pkg;

package body i2c_slave_pkg is

   function f_CTR0  (signal r_register : in STD_LOGIC_VECTOR(31 downto 0))
                                                             return r_CTR0 is
      variable v_return  : r_CTR0;
   begin
      v_return.I2C_OP   := r_register(0);
      v_return.I2C_ADDR := UNSIGNED(r_register( 7 downto  1));
      v_return.BIA      := UNSIGNED(r_register(11 downto  8));
      v_return.BRD      := UNSIGNED(r_register(19 downto 12));
      v_return.BWR      := UNSIGNED(r_register(27 downto 20));
      v_return.x        := (others => '0');
      return v_return;
   end function;

   function f_LT  (signal r_register : in STD_LOGIC_VECTOR(31 downto 0))
                                                             return r_LT is
      variable v_return  : r_LT;
   begin
      v_return.WBCP   := UNSIGNED(r_register(7  downto 0));
      v_return.SCLP   := UNSIGNED(r_register(31 downto 8));
      return v_return;
   end function;

   function f_STD_LOGIC_VECTOR(r_register : in r_LT)
                                                   return STD_LOGIC_VECTOR is
      variable v_return  : STD_LOGIC_VECTOR(31 downto 0);
   begin
      v_return(7  downto 0) := STD_LOGIC_VECTOR(r_register.WBCP);
      v_return(31 downto 8) := STD_LOGIC_VECTOR(r_register.SCLP);
      return v_return;
   end function;

   function f_STD_LOGIC_VECTOR(r_register : in r_CTR0)
                                                   return STD_LOGIC_VECTOR is
      variable v_return  : STD_LOGIC_VECTOR(31 downto 0);
   begin
      v_return(0)            := r_register.I2C_OP;
      v_return( 7 downto  1) := STD_LOGIC_VECTOR(r_register.I2C_ADDR);
      v_return(11 downto  8) := STD_LOGIC_VECTOR(r_register.BIA);
      v_return(19 downto 12) := STD_LOGIC_VECTOR(r_register.BRD);
      v_return(27 downto 20) := STD_LOGIC_VECTOR(r_register.BWR);
      v_return(31 downto 28) := (others => '0');
      return v_return;
   end function;


end i2c_slave_pkg;
