library IEEE;
library work;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use STD.TEXTIO.ALL;
use work.i2c_slave_pkg.ALL;
use work.i2c_tb_pkg.ALL;
use std.textio.ALL;

--! Code intended for simulation only.

entity i2c_master_driver is
   generic(g_WR_DATA_LENGTH : NATURAL := 32;
           g_RD_DATA_LENGTH : NATURAL := 32;
           g_WB_ADDR_LENGTH : NATURAL := 16;
           g_SCL_PERIOD     : TIME    := 250 us);
   port(tb_clk          : in  STD_LOGIC;

        sda_master_i    : in  STD_LOGIC;
        sda_master_o    : out STD_LOGIC;
        scl_master_o    : out STD_LOGIC;

        i2c_addr_op_i   : in  STD_LOGIC_VECTOR(7 downto 0);
        wishbone_addr_i : in  STD_LOGIC_VECTOR(g_WB_ADDR_LENGTH - 1 downto 0);
        wr_data_i       : in  STD_LOGIC_VECTOR(g_WR_DATA_LENGTH - 1 downto 0);
        rd_data_o       : out STD_LOGIC_VECTOR(g_RD_DATA_LENGTH - 1 downto 0);

        start_i         : in  STD_LOGIC;
        start_done_o    : out STD_LOGIC;
        pause_i         : in  STD_LOGIC;
        pause_done_o    : out STD_LOGIC;
        write_i         : in  STD_LOGIC;
        write_done_o    : out STD_LOGIC;
        read_i          : in  STD_LOGIC;
        read_done_o     : out STD_LOGIC);
end i2c_master_driver;

architecture Behavioral of i2c_master_driver is

   type i2c_byte_field is (ADDRESS_0,
                           ADDRESS_1,
                           WISHBONE_HIGH,
                           WISHBONE_LOW,
                           DATA0,
                           DATA1,
                           DATA2,
                           DATA3);

   file     s_file_handler     : TEXT;
   constant c_LOG_PATH         : STRING    := "../test/log/i2c_master_driver.txt";
   constant sep                : CHARACTER := ht;

   signal s_scl_clk       : STD_LOGIC := '1';
   signal s_sda_master_o  : STD_LOGIC := '1';
   signal s_test_id       : NATURAL   := 0;

   signal s_start_done    : STD_LOGIC := '0';
   signal s_pause_done    : STD_LOGIC := '0';
   signal s_write_done    : STD_LOGIC := '0';
   signal s_read_done     : STD_LOGIC := '0';

   signal s_active_link   : STD_LOGIC := '0';

   signal s_i2c_addr_op   : STD_LOGIC_VECTOR(7 downto 0);

   signal s_DTX           : STD_LOGIC_VECTOR(31 downto 0) := (others => '0');

begin

   file_open(s_file_handler, c_LOG_PATH, WRITE_MODE);

   sda_master_o <= s_sda_master_o;
   scl_master_o <= s_scl_clk when s_active_link = '1'
                             else '1';
   start_done_o <= s_start_done;
   pause_done_o <= s_pause_done;
   write_done_o <= s_write_done;
   read_done_o  <= s_read_done;

   scl_clk_process: process
   begin
      s_scl_clk     <= not(s_scl_clk);
        wait for g_SCL_PERIOD/2;
   end process;

   p_sda: process
      variable v_bit_number  : NATURAL := 0;
      variable v_write_data  : STD_LOGIC_VECTOR(g_RD_DATA_LENGTH - 1 downto 0);
      variable v_i2c_addr_op : STD_LOGIC_VECTOR(7 downto 0);

      --! @brief             Function that returns strings for the acked field to be
      --!                    logged.
      --! @param  byte_type  i2c_byte_field type to be translated into a string
      function get_field(byte_type : i2c_byte_field) return STRING is
         variable v_return  : STRING (16 downto 1) := (others => ' ');
      begin
         case byte_type is
            when ADDRESS_0     =>
               v_return := "     [ADDRESS|0]";
            when ADDRESS_1     =>
               v_return := "     [ADDRESS|1]";
            when WISHBONE_HIGH =>
               v_return := "   WISHBONE HIGH";
            when WISHBONE_LOW  =>
               v_return := "    WISHBONE LOW";
            when DATA0         =>
               v_return := "     READ DATA 0";
            when DATA1         =>
               v_return := "     READ DATA 1";
            when DATA2         =>
               v_return := "     READ DATA 2";
            when DATA3         =>
               v_return := "     READ DATA 3";
            when others        =>
               null;
         end case;
         return v_return;
      end function;

      function order_i2c_addr_op(
                  i2c_addr_op : STD_LOGIC_VECTOR(7 downto 0))
                                                    return STD_LOGIC_VECTOR is
         variable v_return  : STD_LOGIC_VECTOR(7 downto 0);
      begin
         for i in 0 to 7 loop
            v_return(7-i) := i2c_addr_op(i);
         end loop;
         return v_return;
      end function;

      --! @brief       Function to reorder data before being sent by driver
      --!              through SDA
      --! @param data  Data to be reordered before being sent
      function order_write_data(
                  data : STD_LOGIC_VECTOR(g_WR_DATA_LENGTH - 1 downto 0))
                                                        return STD_LOGIC_VECTOR is
         variable v_ordered_tmp  : STD_LOGIC_VECTOR(g_WR_DATA_LENGTH - 1 downto 0);
         variable v_ordered      : STD_LOGIC_VECTOR(g_WR_DATA_LENGTH - 1 downto 0);
      begin
         if c_SDA_reverse_bytes = true then
            for i in 0 to g_RD_DATA_LENGTH/8 - 1 loop
               for j in 0 to 7 loop
                  v_ordered_tmp(i*8+j) := data(((g_WR_DATA_LENGTH/8 - 1)-i)*8+j);
               end loop;
            end loop;
         else
            v_ordered := data;
         end if;
         if c_SDA_MSb = true then
            for i in 0 to g_RD_DATA_LENGTH/8 - 1 loop
               for j in 0 to 7 loop
                  v_ordered(i*8+j) := v_ordered_tmp(i*8+(8-j));
               end loop;
            end loop;
         else
            v_ordered := v_ordered_tmp;
         end if;
         return v_ordered;
      end function;

      --! @brief          Procedure to send bytes through SDA
      --! @param byte     Byte to be sent
      procedure send_byte(byte : STD_LOGIC_VECTOR(7 downto 0)) is
      begin
         --! This comes from a falling edge
         for i in 0 to 7 loop
              wait for g_SCL_PERIOD/4;
            s_sda_master_o <= byte(i);
              wait until falling_edge(s_scl_clk);
         end loop;
      end procedure;

      --! @brief             Procedure to check acks
      --! @param test_id     Test ID
      --! @param operation   Operation to be performed, either READ or WRITE
      procedure check_ack(test_id   : NATURAL;
                          operation : STRING (5 downto 1);
                          byte_type : i2c_byte_field) is
         variable v_line_buffer  : LINE;
         variable v_check        : STRING ( 4 downto 1) := (others => ' ');
         variable v_operation    : STRING ( 5 downto 1) := (others => ' ');
         variable v_ack          : STRING (16 downto 1) := (others => ' ');
      begin
           wait until rising_edge(s_scl_clk);
         v_operation          := operation;
         v_ack                := get_field(byte_type);
         if sda_master_i = '1' then
            v_check (4 downto 1) := "Fail";
            write(v_line_buffer,string'(integer'image(test_id)&sep& v_check&sep&v_operation&sep&v_ack));
         else
            v_check (2 downto 1) := "OK";
            write(v_line_buffer,string'(integer'image(test_id)&sep
                                        & v_check             &sep
                                        & v_operation         &sep
                                        & v_ack));
         end if;
         writeline(s_file_handler, v_line_buffer);
         wait until falling_edge(s_scl_clk);
      end procedure;

      --! @brieg Procedure to start/restart I2C communication
      procedure start_restart_I2c is
      begin
           wait until rising_edge(s_scl_clk);
           wait for g_SCL_PERIOD/4;
         s_sda_master_o <= '0';
         s_active_link  <= '1';
           wait until falling_edge(s_scl_clk);
           wait until rising_edge(tb_clk);
         s_start_done <= '1';
           wait until rising_edge(tb_clk);
         s_start_done <= '0';
      end procedure;

      --! @brief Procedure to pause I2c communication
      procedure pause_I2C is
      begin
           wait for g_SCL_PERIOD/4;
         s_sda_master_o   <= '0';
           wait until rising_edge(s_scl_clk);
           wait for g_SCL_PERIOD/4;
         s_sda_master_o   <= '1';
         s_active_link    <= '1';
           --! We allow one clock to let i2c_slave_core.vhd to react
           wait until rising_edge(tb_clk);
         s_pause_done     <= '1';
           wait until rising_edge(tb_clk);
         s_pause_done     <= '0';
         s_active_link    <= '0';
      end procedure;

      procedure read_bit is
         variable v_bit_rcv  : STD_LOGIC;
      begin
           wait until rising_edge(s_scl_clk);
         v_bit_rcv := sda_master_i;
         for i in 0 to (g_SCL_PERIOD/2)/(1 ns) - 1 loop
            if v_bit_rcv /= sda_master_i then
               --! Here we report an error
            end if;
              wait for 1 ns;
         end loop;
           wait until falling_edge(s_scl_clk);
         s_DTX(s_DTX'length - 1) <= v_bit_rcv;
      end procedure;

      procedure shift_bit is
      begin
         for i in 1 to s_DTX'length - 1 loop
            s_DTX(i-1) <= s_DTX(i);
         end loop;
      end procedure;

      procedure read_byte is
      begin
         for i in 0 to 7 loop
            read_bit;
            shift_bit;
         end loop;
      --! At the end we are just in the falling edge of scl
      end procedure;

      procedure place_ack is
      begin
           wait for g_SCL_PERIOD/4;
         s_sda_master_o <= '0';
           wait until falling_edge(s_scl_clk);
         s_sda_master_o <= '1';
      end procedure;


      procedure place_nack is
      begin
           wait for g_SCL_PERIOD/4;
         s_sda_master_o <= '1';
           wait until falling_edge(s_scl_clk);
         s_sda_master_o <= '0';
      end procedure;

   begin

      if start_i = '1' then
         start_restart_I2c;
      elsif pause_i = '1' then
         pause_I2C;
      elsif write_i = '1' then
         s_test_id    <= s_test_id + 1;
         v_write_data := order_write_data(wr_data_i);
         --! 1.- Send [ADDRESS|0]
            v_i2c_addr_op    := i2c_addr_op_i;
            v_i2c_addr_op(0) := '0';
         send_byte(order_i2c_addr_op(v_i2c_addr_op));
         check_ack(s_test_id, "WRITE", ADDRESS_0);
         --! 2.- Send wishbone address high
         send_byte(wishbone_addr_i(15 downto 8));
         check_ack(s_test_id, "WRITE", WISHBONE_HIGH);
         --! 3.- Send wishbone address low
         send_byte(wishbone_addr_i(7 downto 0));
         check_ack(s_test_id, "WRITE", WISHBONE_LOW);
         --! 4.- Send DATA0
         send_byte(wr_data_i(7  downto  0));
         check_ack(s_test_id, "WRITE", DATA0);
         --! 5.- Send DATA1
         send_byte(wr_data_i(15 downto  8));
         check_ack(s_test_id, "WRITE", DATA1);
         --! 6.- Send DATA2
         send_byte(wr_data_i(23 downto 16));
         check_ack(s_test_id, "WRITE", DATA2);
         --! 7.- Send DATA3
         send_byte(wr_data_i(31 downto 24));
         check_ack(s_test_id, "WRITE", DATA3);
         pause_I2C;
         s_write_done <= '1';
           wait until rising_edge(tb_clk);
         s_write_done <= '0';
      elsif read_i  = '1' then
         s_test_id    <= s_test_id + 1;
         v_write_data := order_write_data(wr_data_i);
         --! 1.- Send [ADDRESS|0]
            v_i2c_addr_op    := i2c_addr_op_i;
            v_i2c_addr_op(0) := '0';
         send_byte(order_i2c_addr_op(v_i2c_addr_op));
         check_ack(s_test_id, " READ", ADDRESS_0);
         --! 2.- Send wishbone address high
         send_byte(wishbone_addr_i(15 downto 8));
         check_ack(s_test_id, " READ", WISHBONE_HIGH);
         --! 3.- Send wishbone address low
         send_byte(wishbone_addr_i(7 downto 0));
         check_ack(s_test_id, " READ", WISHBONE_LOW);
         --! 4.- We have to place a start condition
         start_restart_I2c;
         --! 5.- Send [ADDRESS|1]
            v_i2c_addr_op(7 downto 1)    := i2c_addr_op_i(7 downto 1);
            v_i2c_addr_op(0) := '1';
         send_byte(order_i2c_addr_op(v_i2c_addr_op));
         check_ack(s_test_id, " READ", ADDRESS_0);
         --! 6.- Read DATA0
         read_byte;
         place_ack;
         --! 7.- Read DATA1
         read_byte;
         place_ack;
         --! 8.- Read DATA2
         read_byte;
         place_ack;
         --! 9.- Read DATA3
         read_byte;
         place_ack;
         pause_I2C;
         s_read_done <= '1';
           wait until rising_edge(tb_clk);
         s_read_done <= '0';
      end if;
      wait until rising_edge(tb_clk);
   end process;

end;
