library IEEE;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

package multiboot_pkg is

   ----------------------------------------------------------------------------
   -- ICAP configuration data codes
   ----------------------------------------------------------------------------
   --  For custom data codes like the ones for reading back, take a look to:
   --    [OPCODES]        UG380 page 88   table  5-23
   --    [CFG PACKETS]    UG380 page 89   tables 5-{24, 25}
   --    [CG REGISTERS]   UG380 page 90   table  5-30
   ----------------------------------------------------------------------------
   constant DUMMY            : STD_LOGIC_VECTOR(15 downto 0) := X"FFFF";
   constant NOOP             : STD_LOGIC_VECTOR(15 downto 0) := X"2000";
   constant NULL_value       : STD_LOGIC_VECTOR(15 downto 0) := X"1111";
   constant DESYNC           : STD_LOGIC_VECTOR(15 downto 0) := X"000D";
   constant SYNC_H           : STD_LOGIC_VECTOR(15 downto 0) := X"AA99";
   constant SYNC_L           : STD_LOGIC_VECTOR(15 downto 0) := X"5566";
   constant CMD_wr_GEN1      : STD_LOGIC_VECTOR(15 downto 0) := X"3261";
   constant CMD_wr_GEN2      : STD_LOGIC_VECTOR(15 downto 0) := X"3281";
   constant CMD_wr_GEN3      : STD_LOGIC_VECTOR(15 downto 0) := X"32A1";
   constant CMD_wr_GEN4      : STD_LOGIC_VECTOR(15 downto 0) := X"32C1";
   constant CMD_wr_CMD       : STD_LOGIC_VECTOR(15 downto 0) := X"30A1";
   constant CMD_IPROG        : STD_LOGIC_VECTOR(15 downto 0) := X"000E";
   constant CMD_rd_GEN1      : STD_LOGIC_VECTOR(15 downto 0) := X"2A61";
   constant CMD_rd_GEN2      : STD_LOGIC_VECTOR(15 downto 0) := X"2A81";
   constant CMD_rd_GEN3      : STD_LOGIC_VECTOR(15 downto 0) := X"2AA1";
   constant CMD_rd_GEN4      : STD_LOGIC_VECTOR(15 downto 0) := X"2AC1";
   constant CMD_rd_STAT      : STD_LOGIC_VECTOR(15 downto 0) := X"2901";

   attribute a_length : NATURAL;

   -------------------------------------------------------------------
   -- CTR0 register of the control module
   -------------------------------------------------------------------
   -- 8 bits
   -- Wishbone access: Read-write
   -------------------------------------------------------------------
   -- BIT       NAME             Description
   --  3-0       OP               OPeration to be performed
   --  4         PEND             operation PENDing
   --  7-5      x                Reserved
   -------------------------------------------------------------------
   -- List of operations
   -------------------------------------------------------------------
   --  CODE         Description                   Wishbone clocks
   --   4'h0         None
   --   4'h1         Full multiboot
   --   4'h2         Write   MBA register
   --   4'h3         Write   GBA register
   --   4'h4         Refres  MBA register
   --   4'h5         Refre   GBA register
   --   4'h6         Refresh STAT register
   ----------------------------------------------------------------------------
   -- READ operation
   --   ACK is performed automatically
   -- WRITE operation
   --   ACK is performed after [Wishbone clocks] of the specificic operation
   --   ACK is bounded for bad operations (BAD_OP) between [2,16] wb_clks
   ----------------------------------------------------------------------------
   type r_CTR0 is
      record
         OP  : UNSIGNED (3 downto 0);
         PEN : STD_LOGIC;
         x   : STD_LOGIC_VECTOR (7 downto 5);
      end record;

   attribute a_length of r_CTR0 : type is 32;

   constant OP_NONE          : UNSIGNED(3 downto 0) := X"0";
   constant OP_FULLMULTIBOOT : UNSIGNED(3 downto 0) := X"1";
   constant OP_WR_MBA        : UNSIGNED(3 downto 0) := X"2";
   constant OP_WR_GBA        : UNSIGNED(3 downto 0) := X"3";
   constant OP_RD_MBA        : UNSIGNED(3 downto 0) := X"4";
   constant OP_RD_GBA        : UNSIGNED(3 downto 0) := X"5";
   constant OP_RD_STAT       : UNSIGNED(3 downto 0) := X"6";

   -------------------------------------------------------------------
   -- CTR1 register of the control module
   -------------------------------------------------------------------
   -- 8 bits
   -- Wishbone access: Read-only
   -------------------------------------------------------------------
   --     0        OP_ACK           valid OPeration ACKed
   --     1        FSM_RUN          a Finite-State-Machine is RUNning
   --     2        BAD_OP           a BAD OPeration was detected
   --     3        MBAC             MBA register consistency flag
   --     4        GBAC             GBA register consistency flag
   --   7-5        x                Reserved
   -------------------------------------------------------------------
   -- ACK is performed automatically
   -------------------------------------------------------------------
   type r_CTR1 is
      record
         OP_ACK  : STD_LOGIC;
         FSM_RUN : STD_LOGIC;
         BAD_OP  : STD_LOGIC;
         MBAC    : STD_LOGIC;
         GBAC    : STD_LOGIC;
         x       : STD_LOGIC_VECTOR(7 downto 5);
      end record;

   attribute a_length of r_CTR1 : type is 8;

   ------------------------------------------
   -- STAT Spartan 6 configuration register
   ------------------------------------------
   -- 16 bits
   -- Wishbone access: Read-only
   ------------------------------------------
   -- BIT       NAME             Description
   --  0         CRC_ERROR        CRC ERROR
   --  1         ID_ERROR         IDCODE not validated
   --  2         DCM_LOCK         DCMs and PLL locked
   --  3         GTS_CFG_B        Global tristate
   --  4         GWE              Global Write Enable
   --  5         GHIGH_B          GHIGH
   --  6         DEC_ERROR
   --  7         PART_SECURED     Decryption set
   --  8         HSWAPEN          HSWAPEN
   --  11-9      MODE             MODE pins
   --  12        INIT_B           INIT_B
   --  13        DONE             DONE input pin
   --  14        IN_PWRDWN        suspend status
   --  15        SWWD_STRIKEOUT   config error because of invalid sync word
   ------------------------------------------
   -- ACK is performed automatically
   -------------------------------------------------------------------
   type r_STAT is
      record
         CRC     : STD_LOGIC;
         ID      : STD_LOGIC;
         DCM     : STD_LOGIC;
         GTS     : STD_LOGIC;
         GWE     : STD_LOGIC;
         GHIGH_B : STD_LOGIC;
         DEC     : STD_LOGIC;
         SEC     : STD_LOGIC;
         HSWAPEN : STD_LOGIC;
         MODE    : UNSIGNED (11 downto 9);
         INIT_B  : STD_LOGIC;
         DONE    : STD_LOGIC;
         PWRDWN  : STD_LOGIC;
         SOUT    : STD_LOGIC;
      end record;

   attribute a_length of r_STAT : type is 16;

   ------------------------------------------
   -- Bitstream Address Register
   ------------------------------------------
   -- There will be two registers that hold
   -- the addresses of the bitstreams:
   -- + MBA: Multiboot Bitstream Address
   -- + GBA: Golden Bitstream Address
   ------------------------------------------
   type r_BAR is
     record
        ADDR    : UNSIGNED (23 downto 0);
        SPI_OPC : UNSIGNED (31 downto 24);
     end record;

   attribute a_length of r_BAR : type is 32;

   constant c_BAR_default  : r_BAR := (ADDR      => to_unsigned(0, 24),
                                       SPI_OPC   => to_unsigned(0, 8));

   constant c_CTR0_default : r_CTR0 := (OP       => OP_NONE,
                                        PEN      => '0',
                                        x        => (others => '0'));

   constant c_CTR1_default : r_CTR1 := (OP_ACK   => '0',
                                        FSM_RUN  => '0',
                                        BAD_OP   => '0',
                                        MBAC     => '0',
                                        GBAC     => '0',
                                        x        => (others => '0'));

   constant c_STAT_default  : r_STAT := (CRC     => '0',
                                         ID      => '0',
                                         DCM     => '0',
                                         GTS     => '0',
                                         GWE     => '0',
                                         GHIGH_B => '0',
                                         DEC     => '0',
                                         SEC     => '0',
                                         HSWAPEN => '0',
                                         MODE    => (others => '0'),
                                         INIT_B  => '0',
                                         DONE    => '0',
                                         PWRDWN  => '0',
                                         SOUT    => '0');

   constant CTR0_addr       : UNSIGNED (3 downto 0) := X"0";
   constant CTR1_addr       : UNSIGNED (3 downto 0) := X"1";
   constant STAT_addr       : UNSIGNED (3 downto 0) := X"2";
   constant c_MBA_addr      : UNSIGNED (3 downto 0) := X"3";
   constant c_GBA_addr      : UNSIGNED (3 downto 0) := X"4";
   constant c_MBA_ICAP_addr : UNSIGNED (3 downto 0) := X"5";
   constant c_GBA_ICAP_addr : UNSIGNED (3 downto 0) := X"6";

   function f_CTR0 (r_register: STD_LOGIC_VECTOR(r_CTR0'a_length - 1 downto 0))
                                                                 return r_CTR0;
   function f_CTR1 (r_register: STD_LOGIC_VECTOR(r_CTR1'a_length - 1 downto 0))
                                                                 return r_CTR1;
   function f_STAT (r_register: STD_LOGIC_VECTOR(r_STAT'a_length - 1 downto 0))
                                                                 return r_STAT;
   function f_BAR  (r_register: STD_LOGIC_VECTOR(r_BAR'a_length - 1 downto 0))
                                                                  return r_BAR;

   function f_STD_LOGIC_VECTOR (r_register : r_CTR0) return STD_LOGIC_VECTOR;
   function f_STD_LOGIC_VECTOR (r_register : r_CTR1) return STD_LOGIC_VECTOR;
   function f_STD_LOGIC_VECTOR (r_register : r_STAT) return STD_LOGIC_VECTOR;
   function f_STD_LOGIC_VECTOR (r_register : r_BAR)  return STD_LOGIC_VECTOR;

   component multiboot_regs is
   port(wb_rst_i   : in  STD_LOGIC;
        wb_clk     : in  STD_LOGIC;
        wb_we_i    : in  STD_LOGIC;
        wb_stb_i   : in  STD_LOGIC;
        wb_cyc_i   : in  STD_LOGIC;
        wb_sel_i   : in  STD_LOGIC_VECTOR (3  downto 0);
        wb_data_i  : in  STD_LOGIC_VECTOR (31 downto 0);
        wb_data_o  : out STD_LOGIC_VECTOR (31 downto 0);
        wb_addr_i  : in  STD_LOGIC_VECTOR (3  downto 0);
        wb_ack_o   : out STD_LOGIC;
        wb_rty_o   : out STD_LOGIC;
        wb_err_o   : out STD_LOGIC;

        CTR0_o     : out r_CTR0;
        CTR1_i     : in  r_CTR1;
        STAT_i     : in  r_STAT;
        MBA_o      : out r_BAR;
        GBA_o      : out r_BAR;
        MBA_ICAP_i : in  STD_LOGIC_VECTOR (23 downto 0);
        GBA_ICAP_i : in  STD_LOGIC_VECTOR (23 downto 0));
   end component;

   component multiboot_core is
   port(clk         : in  STD_LOGIC;
        rst         : in  STD_LOGIC;

        CTR0_i      : in  r_CTR0;
        CTR1_o      : out r_CTR1;
        STAT_o      : out r_STAT;
        MBA_i       : in  r_BAR;
        GBA_i       : in  r_BAR;
        MBA_ICAP_o  : out STD_LOGIC_VECTOR (23 downto 0);
        GBA_ICAP_o  : out STD_LOGIC_VECTOR (23 downto 0));
   end component;

   procedure order_ICAP(       value :     STD_LOGIC_VECTOR(15 downto 0);
                        signal s_icap: out STD_LOGIC_VECTOR(15 downto 0));


end multiboot_pkg;

package body multiboot_pkg is

   --! @brief            CTR0 register translation function from slv to r_CTR0
   --! @param r_register slv to be translated
   function f_CTR0 (r_register: STD_LOGIC_VECTOR(r_CTR0'a_length - 1 downto 0))
                                                             return r_CTR0 is
      variable v_return  : r_CTR0;
   begin
      v_return.OP  := UNSIGNED(r_register(3 downto 0));
      v_return.PEN := r_register(4);
      v_return.x   := r_register(7 downto 5);
      return v_return;
   end f_CTR0;

   --! @brief            CTR1 register translation function from slv to r_CTR1
   --! @param r_register slv to be translated
   function f_CTR1 (r_register: STD_LOGIC_VECTOR(r_CTR1'a_length - 1 downto 0))
                                                             return r_CTR1 is
      variable v_return  : r_CTR1;
   begin
      v_return.OP_ACK  := r_register(0);
      v_return.FSM_RUN := r_register(1);
      v_return.BAD_OP  := r_register(2);
      v_return.MBAC    := r_register(3);
      v_return.GBAC    := r_register(4);
      v_return.x       := r_register(7 downto 5);
      return v_return;
   end f_CTR1;

   --! @brief            STAT register translation function from slv to r_STAT
   --! @param r_register slv to be translated
   function f_STAT (r_register: STD_LOGIC_VECTOR(r_STAT'a_length - 1 downto 0))
                                                             return r_STAT is
      variable v_return  : r_STAT;
   begin
      v_return.CRC     := r_register(0);
      v_return.ID      := r_register(1);
      v_return.DCM     := r_register(2);
      v_return.GTS     := r_register(3);
      v_return.GWE     := r_register(4);
      v_return.GHIGH_B := r_register(5);
      v_return.DEC     := r_register(6);
      v_return.SEC     := r_register(7);
      v_return.HSWAPEN := r_register(8);
      v_return.MODE    := UNSIGNED(r_register(11 downto 9));
      v_return.INIT_B  := r_register(12);
      v_return.DONE    := r_register(13);
      v_return.PWRDWN  := r_register(14);
      v_return.SOUT    := r_register(15);
      return v_return;
   end f_STAT;

   --! @brief            BAR register translation function from slv to r_BAR
   --! @param r_register slv to be translated
   function f_BAR (r_register: STD_LOGIC_VECTOR(r_BAR'a_length - 1 downto 0))
                                                              return r_BAR is
      variable v_return  : r_BAR;
   begin
      v_return.ADDR    := UNSIGNED(r_register(23 downto  0));
      v_return.SPI_OPC := UNSIGNED(r_register(31 downto 24));
      return v_return;
   end f_BAR;


   --! @brief            CTR0 register translation function from r_CTR0 to slv
   --! @param r_register slv to be translated
   function f_STD_LOGIC_VECTOR (r_register : r_CTR0) return STD_LOGIC_VECTOR is
      variable v_return  : STD_LOGIC_VECTOR(31 downto 0);
   begin
      v_return(3 downto  0) := STD_LOGIC_VECTOR(r_register.OP);
      v_return(4)           := r_register.PEN;
      v_return(7 downto  5) := STD_LOGIC_VECTOR(r_register.x);
      return v_return;
   end function;

   --! @brief            CTR1 register translation function from r_CTR1 to slv
   --! @param r_register slv to be translated
   function f_STD_LOGIC_VECTOR (r_register : r_CTR1) return STD_LOGIC_VECTOR is
      variable v_return  : STD_LOGIC_VECTOR(31 downto 0);
   begin
      v_return(0)           := r_register.OP_ACK;
      v_return(1)           := r_register.FSM_RUN;
      v_return(2)           := r_register.BAD_OP;
      v_return(3)           := r_register.MBAC;
      v_return(4)           := r_register.GBAC;
      v_return(7 downto 5)  := STD_LOGIC_VECTOR(r_register.x);
      return v_return;
   end function;

   --! @brief            STAT register translation function from r_STAT to slv
   --! @param r_register slv to be translated
   function f_STD_LOGIC_VECTOR (r_register : r_STAT) return STD_LOGIC_VECTOR is
      variable v_return  : STD_LOGIC_VECTOR(31 downto 0);
   begin
      v_return(0)           := r_register.CRC;
      v_return(1)           := r_register.ID;
      v_return(2)           := r_register.DCM;
      v_return(3)           := r_register.GTS;
      v_return(4)           := r_register.GWE;
      v_return(5)           := r_register.GHIGH_B;
      v_return(6)           := r_register.DEC;
      v_return(7)           := r_register.SEC;
      v_return(8)           := r_register.HSWAPEN;
      v_return(11 downto 9) := STD_LOGIC_VECTOR(r_register.MODE);
      v_return(12)          := r_register.INIT_B;
      v_return(13)          := r_register.DONE;
      v_return(14)          := r_register.PWRDWN;
      v_return(15)          := r_register.SOUT;
      return v_return;
   end function;

   --! @brief            BAR register translation function from r_BAR to slv
   --! @param r_register slv to be translated
   function f_STD_LOGIC_VECTOR (r_register : r_BAR) return STD_LOGIC_VECTOR is
      variable v_return  : STD_LOGIC_VECTOR(31 downto 0);
   begin
      v_return(23 downto  0) := STD_LOGIC_VECTOR(r_register.ADDR);
      v_return(31 downto 24) := STD_LOGIC_VECTOR(r_register.SPI_OPC);
      return v_return;
   end function;

   procedure order_ICAP(       value :     STD_LOGIC_VECTOR(15 downto 0);
                        signal s_icap: out STD_LOGIC_VECTOR(15 downto 0)) is
   begin
      --! First we order the bits according to UG380 Table2-5 page 37,
      --! as specified in page 126
      s_icap(0)   <= value(7);
      s_icap(1)   <= value(6);
      s_icap(2)   <= value(5);
      s_icap(3)   <= value(4);
      s_icap(4)   <= value(3);
      s_icap(5)   <= value(2);
      s_icap(6)   <= value(1);
      s_icap(7)   <= value(0);
      s_icap(8)   <= value(15);
      s_icap(9)   <= value(14);
      s_icap(10)  <= value(13);
      s_icap(11)  <= value(12);
      s_icap(12)  <= value(11);
      s_icap(13)  <= value(10);
      s_icap(14)  <= value(9);
      s_icap(15)  <= value(8);
   end order_ICAP;

end multiboot_pkg;
