----------------------------------------------------------------------------------
-- Company:        CERN, BE-CO-HT
-- Engineer:       Carlos Gil Soriano
--
-- Create Date:    11:53:14 02/20/2012
-- Design Name:    multiboot module
-- Module Name:    multiboot_regs - Behavioral
-- Project Name:   CONV-TTL-BLO
-- Target Devices: Spartan 6 LX45T
-- Tool versions:  ISE 13.3
-- Description:
--
-- Dependencies:
--
-- Revision        Comment
--  0.01            File created
--  0.1             Added readbacks from ICAP interface
--  0.2             Now there are two CTRL registers: CTR0 and CTR1
--
-- Additional Comments:
--
----------------------------------------------------------------------------------
library IEEE;
library work;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.multiboot_pkg.ALL;

entity multiboot_regs is
   port(wb_rst_i   : in  STD_LOGIC;
        wb_clk     : in  STD_LOGIC;
        wb_we_i    : in  STD_LOGIC;
        wb_stb_i   : in  STD_LOGIC;
        wb_cyc_i   : in  STD_LOGIC;
        wb_sel_i   : in  STD_LOGIC_VECTOR (3  downto 0);
        wb_data_i  : in  STD_LOGIC_VECTOR (31 downto 0);
        wb_data_o  : out STD_LOGIC_VECTOR (31 downto 0);
        wb_addr_i  : in  STD_LOGIC_VECTOR (3  downto 0);
        wb_ack_o   : out STD_LOGIC;
        wb_rty_o   : out STD_LOGIC;
        wb_err_o   : out STD_LOGIC;

        CTR0_o     : out r_CTR0;
        CTR1_i     : in  r_CTR1;
        STAT_i     : in  r_STAT;
        MBA_o      : out r_BAR;
        GBA_o      : out r_BAR;
        MBA_ICAP_i : in  STD_LOGIC_VECTOR (23 downto 0);
        GBA_ICAP_i : in  STD_LOGIC_VECTOR (23 downto 0));
end multiboot_regs;

architecture Behavioral of multiboot_regs is

   signal s_CTR0     : r_CTR0 := c_CTR0_default;
   signal s_CTR1     : r_CTR1;
   signal s_STAT     : r_STAT;

   signal s_MBA      : r_BAR  := c_BAR_default;
   signal s_GBA      : r_BAR  := c_BAR_default;
   signal s_MBA_ICAP : STD_LOGIC_VECTOR(23 downto 0);
   signal s_GBA_ICAP : STD_LOGIC_VECTOR(23 downto 0);

   signal s_wb_ack   : STD_LOGIC := '0';
   signal s_wb_rty   : STD_LOGIC := '0';
   signal s_wb_err   : STD_LOGIC := '0';

begin

   CTR0_o     <= s_CTR0;
   s_CTR1     <= CTR1_i;
   s_STAT     <= STAT_i;

   MBA_o      <= s_MBA;
   GBA_o      <= s_GBA;
   s_MBA_ICAP <= MBA_ICAP_i;
   s_GBA_ICAP <= GBA_ICAP_i;

   wb_ack_o   <= s_wb_ack;
   wb_rty_o   <= s_wb_rty;
   wb_err_o   <= s_wb_err;

   process (wb_clk)
      variable v_wb_addr  : UNSIGNED(3 downto 0) := to_unsigned(0, 4);
   begin
      if rising_edge(wb_clk) then
         v_wb_addr := UNSIGNED(wb_addr_i);
         if wb_rst_i = '1' then
            wb_data_o    <= (others => '0');
            s_wb_ack     <= '0';
            s_wb_rty     <= '0';
            s_wb_err     <= '0';
            wb_data_o    <= (others => '0');
         else
            if    (wb_stb_i = '1' and wb_cyc_i = '1')
               and s_wb_ack = '0'
               and s_wb_rty = '0'
               and s_wb_err = '0' then
               if s_CTR1.FSM_RUN = '0' then
                  s_wb_err <= '1';
               else
                  case wb_we_i is
                     when '0'    =>
                        case v_wb_addr is
                           when CTR0_addr       =>
                              wb_data_o <= f_STD_LOGIC_VECTOR(s_CTR0);
                              s_wb_ack  <= '1';
                           when CTR1_addr       =>
                              wb_data_o <= f_STD_LOGIC_VECTOR(s_CTR1);
                              s_wb_ack  <= '1';
                           when STAT_addr       =>
                              wb_data_o <= f_STD_LOGIC_VECTOR(s_STAT);
                              s_wb_ack  <= '1';
                           when c_MBA_addr      =>
                              wb_data_o <= f_STD_LOGIC_VECTOR(s_MBA);
                              s_wb_ack  <= '1';
                           when c_GBA_addr      =>
                              wb_data_o <= f_STD_LOGIC_VECTOR(s_GBA);
                              s_wb_ack  <= '1';
                           when c_MBA_ICAP_addr =>
                              wb_data_o (31 downto 24) <= (others => '0');
                              wb_data_o (23 downto  0) <= s_MBA_ICAP;
                              s_wb_ack  <= '1';
                           when c_GBA_ICAP_addr =>
                              wb_data_o (31 downto 24) <= (others => '0');
                              wb_data_o (23 downto  0) <= s_GBA_ICAP;
                              s_wb_ack  <= '1';
                           when others =>
                              s_wb_err  <= '1';
                        end case;
                     when others =>
                        case v_wb_addr is
                           when CTR0_addr       =>
                              s_CTR0    <= f_CTR0(wb_data_i);
                              s_wb_ack  <= '1';
                           when c_MBA_addr      =>
                              s_MBA     <= f_BAR(wb_data_i);
                              s_wb_ack  <= '1';
                           when c_GBA_addr      =>
                              s_GBA     <= f_BAR(wb_data_i);
                              s_wb_ack  <= '1';
                           when others =>
                              s_wb_err  <= '1';
                        end case;
                  end case;
               end if;
            else
            end if;
         end if;
      else
      end if;
   end process;

end Behavioral;
