-- TestBench Template
library IEEE;
library work;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use STD.TEXTIO.ALL;
use work.multiboot_pkg.ALL;


entity multiboot_core_tb is
   end multiboot_core_tb;

architecture behavior of multiboot_core_tb is

   constant c_CLK_PERIOD : TIME := 50 ns;

   signal rst           : STD_LOGIC;
   signal clk           : STD_LOGIC;
   signal CTR0_o        : r_CTR0;
   signal CTR1_i        : r_CTR1;
   signal STAT_i        : r_STAT;
   signal MBA_i         : r_BAR;
   signal GBA_i         : r_BAR;
   signal MBA_ICAP_o    : STD_LOGIC_VECTOR (23 downto 0);
   signal GBA_ICAP_o    : STD_LOGIC_VECTOR (23 downto 0));

   file file_handler    : TEXT;
   constant c_LOG_PATH  : STRING := "../test/log/multiboot_core_tb.log";
   constant sep         : CHARACTER := ht;

begin

   uut: multiboot_core
   port map(clk         => clk,
            rst         => rst,
            CTR0_o      => CTR0_o,
            CTR1_i      => CTR1_i,
            STAT_i      => STAT_i,
            MBA_i       => MBA_i,
            GBA_i       => GBA_i,
            MBA_ICAP_o  => MBA_addr_o,
            GBA_ICAP_o  => GBA_addr_o);

   clk_process: process
   begin
      clk <= '1';
      wait for c_CLK_PERIOD/2;
      clk <= '0';
      wait for c_CLK_PERIOD/2;
   end process;


   tb : process

      procedure reset_proc is
      begin
           wait until rising_edge(clk);
         rst       <= '1';
         CTR1_i   <= c_CTR1_default;
           wait for c_CLK_PERIOD;
         rst       <= '0';
           wait for c_CLK_PERIOD*25;
      end reset_proc;

      procedure assert16_to_file(data :       STD_LOGIC_VECTOR(15 downto 0);
                                 reference :  STD_LOGIC_VECTOR(15 downto 0);
                                 test_id :    STRING;
                                 msg :        STRING) is
         variable line_buffer : LINE;
      begin
         if data = reference then
            write(line_buffer, string'(test_id&sep&"OK"&sep&msg));
         else
            write(line_buffer, string'(test_id&sep&"Fail"&sep&msg));
         end if;
         writeline(file_handler, line_buffer);
         assert data = reference  report "Bad"&msg;
      end procedure;

      procedure assert32_to_file(data :       STD_LOGIC_VECTOR(31 downto 0);
                                 reference :  STD_LOGIC_VECTOR(31 downto 0);
                                 test_id :    STRING;
                                 msg :        STRING) is
         variable line_buffer : LINE;
      begin
         if data = reference then
            write(line_buffer, string'(test_id&sep&"OK"&sep&msg));
         else
            write(line_buffer, string'(test_id&sep&"Fail"&sep&msg));
         end if;
         writeline(file_handler, line_buffer);
         assert data = reference  report "Bad"&msg;
      end procedure;

      variable header_line_buffer : LINE;

   begin
      file_open(file_handler, c_LOG_PATH, WRITE_MODE);
        write(header_line_buffer, string'("-----------------------------------------"));
      writeline(file_handler, header_line_buffer);
        write(header_line_buffer, string'("-----     multiboot_core_tb.vhd     -----"));
      writeline(file_handler, header_line_buffer);
        write(header_line_buffer, string'("-----------------------------------------"));
      writeline(file_handler, header_line_buffer);
        write(header_line_buffer, string'(""));
      writeline(file_handler, header_line_buffer);
        write(header_line_buffer, string'("Test ID"&sep&"Status"&sep&"Description"));
      writeline(file_handler, header_line_buffer);


      -- Initialization assert
        wait for 2*c_CLK_PERIOD;
      reset_proc;
         assert16_to_file(GENERAL1_ICAP_o, X"0000", "0x0", "GENERAL1_ICAP register initialization");
         assert16_to_file(GENERAL2_ICAP_o, X"0000", "0x1", "GENERAL2_ICAP register initialization");
         assert16_to_file(GENERAL3_ICAP_o, X"0000", "0x2", "GENERAL3_ICAP register initialization");
         assert16_to_file(GENERAL4_ICAP_o, X"0000", "0x3", "GENERAL4_ICAP register initialization");
         assert16_to_file(STAT_ICAP_o, X"0000", "0x4", "STAT_ICAP register initialization");

      GENERAL1_i     <= X"1234";
      GENERAL2_i     <= X"5678";
      GENERAL3_i     <= X"9ABC";
      GENERAL4_i     <= X"DEF0";

      -- We will test the full multiboot sequence
      CTR0_i(3 downto 0)    <= OP_FULLMULTIBOOT;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 13*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0x5", "Multiboot sequence is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_WR_GENERAL1;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 5*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0x6", "GENERAL1 write sequence is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_WR_GENERAL2;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 5*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0x7", "GENERAL2 write sequence is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_WR_GENERAL3;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 5*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0x8", "GENERAL3 write sequence is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_WR_GENERAL4;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 5*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0x9", "GENERAL4 write sequence is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_RD_GENERAL1;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 15*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0xA", "GENERAL1 fsm readback is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_RD_GENERAL2;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 15*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0xB", "GENERAL2 fsm readback is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_RD_GENERAL3;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 15*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0xC", "GENERAL3 fsm readback is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      CTR0_i(3 downto 0)    <= OP_RD_GENERAL4;
      CTR0_i(4)             <= '1';
      CTR0_i(15 downto 5)   <= (others => '0');
      -- Every wb_clk cycle the fsm should move one state. The chain state will be:
      -- {S0_IDLE, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11, S12, Q0_OK, }
        wait for 15*c_CLK_PERIOD;
        wait for c_CLK_PERIOD/2;
        wait until rising_edge(clk);
      assert16_to_file(CTR1_o, X"0001", "0xD", "GENERAL4 fsm readback is correct");
      -- We act as the multiboot_regs.vhd by deasserting CTR0_i(4) bit
      CTR0_i(4)             <= '0';
        wait for 100*c_CLK_PERIOD;

      assert false report "No error. Simulation ends." severity failure;
   end process tb;

end;
